@@ -1,12 +1,13 @@
-from .Typing import Callable, Type, Any, Union, Tuple
+from .Typing import Callable, Any, Union
 import functools
 import threading
 from .Functions import areoneof, isoneof, isoneof_strict, isoftype
 from .Exceptions import OverloadDuplication, OverloadNotFound, ValidationTypeError, ValidationValueError, ValidationReturnTypeError, TimeoutError
 __validation_set = set()
+__validation_instantiation_rule = dict()
 
 
-def __validate_type(func: Callable, v: Any, T: Type, validation_func: Callable[[Any], bool] = isoftype, msg: str = None) -> None:
+def __validate_type(func: Callable, v: Any, T: type, validation_func: Callable[[Any], bool] = isoftype, msg: str = None) -> None:
     if not validation_func(v, T):
         raise ValidationTypeError(
             msg or f"In {func.__module__}.{func.__qualname__}(...)\nThe argument is: '{ v.__qualname__ if hasattr(v, '__qualname__') else v}'\nIt has the type of '{type(v)}'\nIt is marked as type(s): '{T}'")
@@ -19,16 +20,16 @@ def __validate_condition(func: Callable, v: Any, constraint: Callable[[Any], boo
 
 
 def __validate_arg(func: Callable, curr_arg: Any, curr_inner_arg: Any) -> None:
-    if isoneof(curr_arg, [list, Tuple]):
+    if isoneof(curr_arg, [list, tuple]):
         # multiple type only:
-        if areoneof(curr_arg, [Type]):
+        if areoneof(curr_arg, [type]):
             __validate_type(func, curr_inner_arg, curr_arg, isoneof)
 
         else:  # maybe with condition:
             class_type, constraint = curr_arg[0], curr_arg[1]
 
             # Type validation
-            if isoneof(class_type, [list, Tuple]):
+            if isoneof(class_type, [list, tuple]):
                 __validate_type(func, curr_inner_arg, class_type, isoneof)
             else:
                 __validate_type(func, curr_inner_arg, class_type, isinstance)
@@ -41,7 +42,7 @@ def __validate_arg(func: Callable, curr_arg: Any, curr_inner_arg: Any) -> None:
         __validate_type(func, curr_inner_arg, curr_arg)
 
 
-def validate(*args, return_type=None) -> Callable:
+def validate(*args, return_type=None, can_instantiate_multiple_times: bool = False) -> Callable:
     """validate decorator
 
         Is passed types of variables to perform type checking over\n
@@ -59,17 +60,26 @@ def validate(*args, return_type=None) -> Callable:
     """
     from .Exceptions import ValidationDuplicationError, ValidationTypeError, ValidationValueError, ValidationReturnTypeError
 
-    def wrapper(func: Callable) -> Callable:
-        global __validation_set
+    def deco(func: Callable) -> Callable:
+        if not isinstance(func, Callable):
+            raise ValueError("validate decorator must decorate a callable")
+        global __validation_set, __validation_instantiation_rule
         func_id = f"{func.__module__}.{func.__qualname__}"
+
+        if func_id not in __validation_instantiation_rule:
+            __validation_instantiation_rule[func_id] = can_instantiate_multiple_times
+        assert can_instantiate_multiple_times == __validation_instantiation_rule[
+            func_id], "can't change instantiation status on runtime"
+
         if func_id not in __validation_set:
             __validation_set.add(func_id)
         else:
-            raise ValidationDuplicationError(
-                "validate decorator is being used on two functions in the same module with the same name\nmaybe use @overload instead")
+            if not __validation_instantiation_rule[func_id]:
+                raise ValidationDuplicationError(
+                    "validate decorator is being used on two functions in the same module with the same name\nmaybe use @overload instead")
 
         @ functools.wraps(func)
-        def inner(*inner_args, **inner_kwargs) -> Any:
+        def wrapper(*inner_args, **inner_kwargs) -> Any:
             for i in range(min(len(args), len(inner_args))):
                 if args[i] is not None:
                     __validate_arg(func, args[i], inner_args[i])
@@ -78,8 +88,8 @@ def validate(*args, return_type=None) -> Callable:
                 msg = f"In {func.__module__}.{func.__qualname__}(...)\nThe returned value is: '{ res.__qualname__ if hasattr(res, '__qualname__') else res}'\nIt has the type of '{type(res)}'\nIt is marked as type(s): '{return_type}'"
                 __validate_type(func, res, return_type, msg=msg)
             return res
-        return inner
-    return wrapper
+        return wrapper
+    return deco
 
 
 @ validate(Callable)
@@ -120,7 +130,7 @@ def memo(func: Callable) -> Callable:
     Args:
         func (Callable): function to memorize
     """
-    cache: dict[Tuple, Any] = {}
+    cache: dict[tuple, Any] = {}
 
     @ functools.wraps(func)
     def wrapper(*args, **kwargs):
@@ -130,7 +140,7 @@ def memo(func: Callable) -> Callable:
     return wrapper
 
 
-__overload_dict: dict[str, dict[Tuple, Callable]] = dict()
+__overload_dict: dict[str, dict[tuple, Callable]] = dict()
 
 
 def overload(*types) -> Callable:
@@ -143,7 +153,11 @@ def overload(*types) -> Callable:
     def print_info(name,age):
         ...\n\n
 
-    also, use 'None' to skip\n\n\n
+    * use None to skip argument
+    * use no arguments to mark as default function
+    * you should overload in decreasing order of specificity! e.g @overload(int) should appear in the code before @overload(Any)
+
+    \n\n\n
     \nRaises:
         OverloadDuplication: if a functions is overloaded twice (or more) with same argument types
         OverloadNotFound: if an overloaded function is called with types that has no variant of the function
@@ -155,15 +169,16 @@ def overload(*types) -> Callable:
     if len(types) == 1 and type(types[0]).__name__ == "function":
         raise ValueError("can't create an overload without defining types")
     global __overload_dict
-    # allow input of both tuples and lists for flexibility
-    types = list(types)
-    for i, maybe_list_of_types in enumerate(types):
-        if isoneof(maybe_list_of_types, [list, Tuple]):
-            types[i] = tuple(sorted(list(maybe_list_of_types),
-                                    key=lambda sub_type: sub_type.__name__))
-    types = tuple(types)
+    # allow input of both tuples and lists for flexibly
+    if len(types) > 0:
+        types = list(types)
+        for i, maybe_list_of_types in enumerate(types):
+            if isoneof(maybe_list_of_types, [list, tuple]):
+                types[i] = tuple(sorted(list(maybe_list_of_types),
+                                        key=lambda sub_type: sub_type.__name__))
+        types = tuple(types)
 
-    def wrapper(func: Callable) -> Callable:
+    def deco(func: Callable) -> Callable:
         if not isinstance(func, Callable):
             raise TypeError("overload decorator must be used on a callable")
 
@@ -181,16 +196,24 @@ def overload(*types) -> Callable:
         __overload_dict[name][types] = func
 
         @ functools.wraps(func)
-        def inner(*args, **kwargs) -> Any:
-
+        def wrapper(*args, **kwargs) -> Any:
+            default_func = None
             # select correct overload
             for variable_types, curr_func in __overload_dict[f"{func.__module__}.{func.__qualname__}"].items():
+                if len(variable_types) == 0:
+                    if default_func is None:
+                        default_func = curr_func
+                        continue
+                    else:
+                        # will not reach here because of duplicate overloading so this is redundant
+                        raise ValueError("Can't have two default functions")
+
                 if len(variable_types) != len(args):
                     continue
 
                 for i, variable_type in enumerate(variable_types):
                     if variable_type is not None:
-                        if isoneof(variable_type, [list, Tuple]):
+                        if isoneof(variable_type, [list, tuple]):
                             if not isoneof_strict(args[i], variable_type):
                                 break
                         else:
@@ -199,12 +222,14 @@ def overload(*types) -> Callable:
                 else:
                     return curr_func(*args, **kwargs)
 
+            if default_func is not None:
+                return default_func(*args, **kwargs)
             # or raise exception if no overload exists for current arguments
             raise OverloadNotFound(
                 f"function {func.__module__}.{func.__qualname__} is not overloaded with {[type(v) for v in args]}")
 
-        return inner
-    return wrapper
+        return wrapper
+    return deco
 
 
 @ validate(Callable)
@@ -265,25 +290,22 @@ def deprecate(obj: Union[str, Callable] = None) -> Callable:
     # if callable(obj):
     if isinstance(obj, Callable):
         @ functools.wraps(obj)
-        def inner(*args, **kwargs) -> Any:
+        def wrapper(*args, **kwargs) -> Any:
             warning(
                 f"As marked by the developer, {obj.__module__}.{obj.__qualname__} is deprecated")
             return obj(*args, **kwargs)
-        return inner
+        return wrapper
 
-    def wrapper(func: Callable) -> Callable:
+    def deco(func: Callable) -> Callable:
         @ functools.wraps(func)
-        def inner(*args, **kwargs) -> Any:
+        def wrapper(*args, **kwargs) -> Any:
             warning(
                 f"As marked by the developer, {func.__module__}.{func.__qualname__} is deprecated")
             if obj:
                 print(obj)
             return func(*args, **kwargs)
-        return inner
-    return wrapper
-
-
-# @PartiallyImplemented
+        return wrapper
+    return deco
 
 
 @ validate(Callable)
@@ -312,9 +334,9 @@ def limit_recursion(max_depth: int, return_value=None, quiet: bool = True):
     import re
     from .Colors import warning
 
-    def wrapper(func):
+    def deco(func):
         @ functools.wraps(func)
-        def inner(*args, **kwargs):
+        def wrapper(*args, **kwargs):
             depth = functools.reduce(
                 lambda count, line:
                     count + 1 if re.search(f"{func.__name__}\(.*\)$", line)
@@ -329,37 +351,72 @@ def limit_recursion(max_depth: int, return_value=None, quiet: bool = True):
                     return return_value
                 return args, kwargs
             return func(*args, **kwargs)
-        return inner
-    return wrapper
+        return wrapper
+    return deco
 
 
-def timeout(timeout: float):
+@validate([int, float])
+def timeout(timeout: int | float) -> Callable:
+    """A decorator to limit runtime for a function
+
+    Args:
+        timeout (int | float): allowed runtime duration
+
+    Raises:
+        thread_error: if there is a thread related error
+        function_error: if there is an error in the decorated function
+
+    Returns:
+        Callable: The decorated function
+    """
     # https://stackoverflow.com/a/21861599/6416556
-    def deco(func):
+    def timeout_deco(func: Callable) -> Callable:
+        if not isinstance(func, Callable):
+            raise ValueError("timeout must decorate a function")
+
         @functools.wraps(func)
         def wrapper(*args, **kwargs):
             res = [
                 TimeoutError(f'{func.__module__}.{func.__qualname__} timed out after {timeout} seconds!')]
 
-            def newFunc():
+            def timeout_wrapper() -> None:
                 try:
                     res[0] = func(*args, **kwargs)
                 except Exception as function_error:
                     res[0] = function_error
 
-            t = threading.Thread(target=newFunc)
-            t.daemon = True
+            t = threading.Thread(target=timeout_wrapper, daemon=True)
             try:
                 t.start()
                 t.join(timeout)
             except Exception as thread_error:
                 raise thread_error
-            ret = res[0]
-            if isinstance(ret, BaseException):
-                raise ret
-            return ret
+            if isinstance(res[0], BaseException):
+                raise res[0]
+            return res[0]
         return wrapper
-    return deco
+    return timeout_deco
+
+
+@validate(Callable, Callable)
+def attach(before: Callable = None, after: Callable = None) -> Callable:
+    if before is None and after is None:
+        raise ValueError("You must supply at least one function")
+
+    def attach_deco(func: Callable):
+        if not isinstance(func, Callable):
+            raise ValueError("attach must decorate a function")
+
+        @functools.wraps(func)
+        def wrapper(*args, **kwargs):
+            if before is not None:
+                before()
+            res = func(*args, **kwargs)
+            if after is not None:
+                after()
+            return res
+        return wrapper
+    return attach_deco
 
 
 __all__ = [
@@ -375,5 +432,6 @@ __all__ = [
     "deprecate",
     "atomic",
     "limit_recursion",
-    "timeout"
+    "timeout",
+    "attach"
 ]
