@@ -1,7 +1,9 @@
 # -*- coding: utf-8 -*-
+import shutil
 import os
 from .Decorators import validate
 from typing import Union
+from pathlib import Path
 
 
 @validate(str, list)
@@ -172,10 +174,16 @@ def delete_directory(path: str) -> None:
         path (str): _description_
     """
     if is_directory(path):
-        for file in get_files(dir):
-            delete_file(f"{path}\\{file}")
-        for dir in get_directories(path):
-            delete_directory(f"{path}\\{dir}")
+        clear_directory(path)
+        os.rmdir(path)
+
+
+@validate(str)
+def clear_directory(path: str) -> None:
+    for file in get_files(path):
+        delete_file(f"{path}\\{file}")
+    for dir in get_directories(path):
+        delete_directory(f"{path}\\{dir}")
 
 
 @validate(str)
@@ -191,7 +199,6 @@ def create_directory(path: str) -> None:
 
 @validate(str, str)
 def get_file_type_from_directory(path: str, file_type: str) -> list[str]:
-    from pathlib import Path
     return list(
         filter(
             lambda name: Path(f"{path}\\{name}").suffix == file_type,
@@ -202,7 +209,6 @@ def get_file_type_from_directory(path: str, file_type: str) -> list[str]:
 
 @validate(str, str)
 def get_file_type_from_directory_recursively(path: str, file_type: str):
-    from pathlib import Path
     res = []
     for dir in get_directories(path):
         res.extend(f"{dir}\\{v}" for v in get_file_type_from_directory_recursively(
@@ -216,6 +222,34 @@ def get_file_type_from_directory_recursively(path: str, file_type: str):
     return res
 
 
+def rename_file(path: str, new_name: str) -> None:
+    new_path = "./" + \
+        "/".join(Path(path).parts[:-1])+"/"+new_name+Path(path).suffix
+    move_file(path, new_path)
+
+
+def move_file(old_path: str, new_path: str) -> None:
+    os.rename(old_path, new_path)
+
+
+async def open_file(file_path: str, application_path: str):
+    import subprocess
+    p = subprocess.Popen([application_path, file_path])
+    return_code = p.wait()
+
+
+def move_directory(old_path: str, new_path: str) -> None:
+    shutil.move(old_path, new_path)
+
+
+def copy_file(src: str, dest: str) -> None:
+    shutil.copy(src, dest)
+
+
+def copy_directory(src: str, dest: str) -> None:
+    shutil.copy(src, dest)
+
+
 __all__ = [
     "write_to_file",
     "path_exists",
@@ -229,7 +263,14 @@ __all__ = [
     "get_files_and_directories",
     "get_directories",
     "delete_directory",
+    "clear_directory",
     "create_directory",
     "get_file_type_from_directory",
-    "get_file_type_from_directory_recursively"
+    "get_file_type_from_directory_recursively",
+    "rename_file",
+    "move_file",
+    "open_file",
+    "move_directory",
+    "copy_file",
+    "copy_directory"
 ]
